from typing import List
from collections import namedtuple

from serial import Serial


class SerialInterface:
    """ Wrapper around Python's own serial package """
    _default_config: dict = {'port': '/dev/ttyACM0', 'baudrate': 115200, 'term_seq': b'uart:~$ ', 'timeout': 1}

    def __init__(self, config: dict = None) -> None:
        if config is None:
            config = self._default_config
        self.config = namedtuple('config', config.keys())(**config)

    def execute(self, cmd: str) -> List[str]:
        """ Execute a cmd synchronosly. The will halt and wait untill the termination char is received """
        print(f"Executing: {cmd!r}")
        _cmd = self._add_line_ending(line=cmd)
        rx_buffer = []
        with Serial(port=self.config.port, baudrate=self.config.baudrate, timeout=self.config.timeout) as serial:
            serial.write(_cmd.encode())
            while True:
                rx_data = serial.readline()
                if rx_data == self.config.term_seq.encode():
                    break
                line_buffer = rx_data.decode('ascii').rstrip()
                if line_buffer != "":
                    rx_buffer.append(line_buffer)
        return rx_buffer

    def write_cmd(self, cmd: str) -> None:
        """ Write a single command out onto serial """
        print(f"Writing cmd: {cmd!r}")
        _cmd = self._add_line_ending(line=cmd)
        with Serial(port=self.config.port, baudrate=self.config.baudrate, timeout=self.config.timeout) as serial:
            serial.write(_cmd.encode())

    def read_line(self) -> str:
        """ Read one line at the time """
        with Serial(port=self.config.port, baudrate=self.config.baudrate, timeout=self.config.timeout) as serial:
            rx_data = serial.readline()
        return rx_data.decode('ascii').rstrip()

    def read_until_empty_line(self) -> str:
        running = True
        with Serial(port=self.config.port, baudrate=self.config.baudrate, timeout=5) as serial:
            while running:
                rx_buffer = serial.readline()
                if rx_buffer == b'':
                    running = False
                else:
                    yield rx_buffer.decode('ascii').rstrip()

    @staticmethod
    def _add_line_ending(line: str) -> str:
        if line[-1] != '\r':
            line += '\r'
        return line



from collections import namedtuple

from pigpio import pi


class GpioInterface:
    _default_config: dict = {'host': 'localhost', 'port': 8888}

    def __init__(self, config: dict = None) -> None:
        if config is None:
            config = self._default_config
        self.config = namedtuple('config', config.keys())(**config)
        self.gpio_pi = pi(host=self.config.host, port=self.config.port)

    def __del__(self) -> None:
        self.gpio_pi.stop()

    def write(self, pin: int, state: bool) -> None:
        self.gpio_pi.write(gpio=pin, level=state)

    def read(self, pin: int) -> int:
        return self.gpio_pi.read(gpio=pin)



from collections import namedtuple
from time import sleep

from .serial_wrapper import SerialInterface
from .gpio_wrapper import GpioInterface


class HardwareControler:
    def __init__(self, serial_config: dict, gpio_config: dict, pinout: dict):
        self.pinout = namedtuple('pinout', pinout.keys())(**pinout)
        self._serial = SerialInterface(config=serial_config)
        self._gpio = GpioInterface(config=gpio_config)

    def setup_wireless_adapter(self):
        """ Setting up the shell for automatic actions """
        self._serial.execute(cmd="shell colors off")
        self._serial.execute(cmd="shell echo off")

    def trigger_reset(self, period: int):
        """ Blocking reset action, setting the reset pin high for period seconds"""
        self._gpio.write(self.pinout.reset, True)
        sleep(period)
        self._gpio.write(self.pinout.reset, False)

    def power_device_under_test(self, power_state: bool):
        """ Turns the device under test on/off """
        self._gpio.write(self.pinout.power, power_state)

    def enable_bluetooth(self):
        """ Prepars the wireless adapter for BT actions """
        self._serial.execute(cmd="bt enable")

    def sync_serial(self, cmd: str) -> list:
        """ Executes a serial command in blocking fashion, waiting until the termincation char is reached """
        return self._serial.execute(cmd=cmd)

    def async_serial(self, cmd: str):
        pass

    def gpio_write(self, pin: int, state: bool) -> None:
        self._gpio.write(pin=pin, state=state)

    def gpio_read(self, pin: int) -> int:
        pass

